import pytest


def test(level=0):
    """
    Execute openbtmixing's full test suite.

    :param level: Level 0 indicates the least logging.  Level 2 indicates the
        maximum logging possible.
    """
    VERBOSITY = [0, 1, 2]

    args = []
    if level not in VERBOSITY:
        raise ValueError(f"level must be in {VERBOSITY}")
    elif level >= 1:
        args = ["-" + "v"*level]

    return (pytest.main(args + ["--pyargs", "openbtmixing.tests"]) == 0)



'''
Interface helpers
'''

import numpy as np
import sys
import subprocess
import tempfile
import shutil
import os
import typing
import numbers

from scipy.stats import norm
from pathlib import Path
from scipy.stats import spearmanr

def read_in_preds(fpath, modelname, nummodels, q_upper, q_lower, readmean, readstd, readwts):
    # Init containers
    mdraws = []
    sdraws = []
    wdraws = {}
    out = {}
    if readmean:
        mdraw_files = sorted(list(fpath.glob(modelname + ".mdraws*")))
        
        for f in mdraw_files:
            read = open(f, "r")
            lines = read.readlines()
            if lines[0] != '\n' and lines[1] != '\n':  
                mdraws.append(np.loadtxt(f))

        # print(mdraws[0].shape); print(len(mdraws))
        mdraws = np.concatenate(mdraws, axis=1)  # Got rid of the transpose
        n_test = len(mdraws[0])
        
        pred_mean = np.empty(n_test)
        pred_sd = np.empty(n_test)
        pred_med = np.empty(n_test)
        pred_lower = np.empty(n_test)
        pred_upper = np.empty(n_test)
    
    if readstd:
        sdraw_files = sorted(list(fpath.glob(modelname + ".sdraws*")))
    
        for f in sdraw_files:
            read = open(f, "r")
            lines = read.readlines()
            if lines[0] != '\n' and lines[1] != '\n':  # If it's nonempty
                sdraws.append(np.loadtxt(f))
        # print(sdraws[0]); print(sdraws[0][0])
        # print(len(sdraws)); print(len(sdraws[0])); print(len(sdraws[0][0]))
        sdraws = np.concatenate(sdraws, axis=1)  # Got rid of the transpose

        n_test = len(sdraws[0])
        sigma_mean = np.empty(n_test)
        sigma_sd = np.empty(n_test)
        sigma_med = np.empty(n_test)
        sigma_lower = np.empty(n_test)
        sigma_upper = np.empty(n_test)

    # Initialize the wdraws dictionary
    if readwts:
        # Get the weight files
        for k in range(nummodels):
            wdraw_files = sorted(list(fpath.glob(modelname + ".w" + str(k + 1) + "draws*")))
            wdraws_temp = []
            for f in wdraw_files:
                read = open(f, "r")
                lines = read.readlines()
                if lines[0] != '\n' and lines[1] != '\n':  # If it's nonempty
                    wdraws_temp.append(np.loadtxt(f))

            # Store the wdraws array in the self.wdraws dictionary under the
            # key wtname
            wtname = "w" + str(k + 1)
            wdraws[wtname] = np.concatenate(wdraws_temp, axis=1) 

        # Initialize summary statistic matrices for the wts
        n_test = len(wdraws[wtname][0])
        wts_mean = np.empty((n_test, nummodels))
        wts_sd = np.empty((n_test, nummodels))
        wts_med = np.empty((n_test, nummodels))
        wts_lower = np.empty((n_test, nummodels))
        wts_upper = np.empty((n_test, nummodels))


    for j in range(n_test):
        if readmean:
            pred_mean[j] = np.mean(mdraws[:, j])
            pred_sd[j] = np.std(mdraws[:, j], ddof=1)
            pred_med[j] = np.quantile(mdraws[:, j], 0.50)
            pred_lower[j] = np.quantile(mdraws[:, j], q_lower)
            pred_upper[j] = np.quantile(mdraws[:, j], q_upper)

        if readstd:
            sigma_mean[j] = np.mean(sdraws[:, j])
            sigma_sd[j] = np.std(sdraws[:, j], ddof=1)
            sigma_med[j] = np.quantile(sdraws[:, j], 0.50)
            sigma_lower[j] = np.quantile(sdraws[:, j], q_lower)
            sigma_upper[j] = np.quantile(sdraws[:, j], q_upper)

        if readwts:
            for k in range(nummodels):
                wtname = "w" + str(k + 1)
                wts_mean[j][k] = np.mean(wdraws[wtname][:, j])
                wts_sd[j][k] = np.std(wdraws[wtname][:, j], ddof=1)
                wts_med[j][k] = np.quantile(wdraws[wtname][:, j], 0.50)
                wts_lower[j][k] = np.quantile(wdraws[wtname][:, j],q_lower)
                wts_upper[j][k] = np.quantile(wdraws[wtname][:, j], q_upper)
            
    if readmean:
        out_pred = {"draws":mdraws,"mean":pred_mean,"sd":pred_sd,
                    "med":pred_med,"lb":pred_lower,"ub":pred_upper}
        out["pred"] = out_pred

    if readstd:
        out_sigma = {"draws":sdraws,"mean":sigma_mean,"sd":sigma_sd,
                "med":sigma_med,"lb":sigma_lower,"ub":sigma_upper}
        out["sigma"] = out_sigma

    if readwts:
        out_wts = {"draws":wdraws,"mean":wts_mean,"sd":wts_sd,
                        "med":wts_med,"lb":wts_lower,"ub":wts_upper}
        out["wts"] = out_wts

    return out


def run_model(fpath, tc, cmd="openbtcli",
              local_openbt_path="", google_colab=False):
    """
    Private function, run the cpp program via the command line using
    a subprocess.

    This is hardcoded to use the OpenBT command line tools installed with the
    package.  It requires that mpirun be in the PATH and that the mpirun that it
    finds should be from an MPI implementation compatible with the MPI
    implementation that was used to build the OpenBT C++ command line tools.

    .. todo::
        * What error checking should be done for fpath?
        * Remove local_openbt_path and google_colab.
    """
    PKG_ROOT = Path(__file__).parent.resolve()
    BIN_PATH = PKG_ROOT.joinpath("bin")
    # This should be matched to the CLTs installed in the package as specified
    # in setup.py.
    KNOWN_COMMANDS = {"openbtcli", "openbtpred", "openbtmixingwts"}

    # ----- ERROR CHECK ARGUMENTS
    cmd_with_path = BIN_PATH.joinpath(cmd)

    if (not isinstance(fpath, str)) and (not isinstance(fpath, Path)):
        raise TypeError(f"fpath is not a string or Path object ({fpath})")
    elif not isinstance(tc, numbers.Integral):
        raise TypeError(f"tc is not an integer ({tc})")
    elif tc <= 0:
        raise ValueError(f"tc is not a positive integer ({tc})")
    elif cmd not in KNOWN_COMMANDS:
        raise ValueError(f"Unknown OpenBT command line tool {cmd}")
    elif not cmd_with_path.is_file():
        raise RuntimeError(f"Could not locate {cmd} in package installation")
    elif local_openbt_path != "":
        raise NotImplementedError("local_openbt_path not supported")
    elif google_colab:
        raise NotImplementedError("google_colab not supported")

    # ----- ERROR CHECK MPI SETUP
    if shutil.which("mpirun") is None:
        msg = "Add to PATH the folder that contains the mpirun\n"
        msg += "of the MPI implementation used to build OpenBT CLTs"
        raise RuntimeError(msg)

    try:
        subprocess.run(["mpirun", "-np", str(tc),
                       str(cmd_with_path), str(fpath)],
                       stdin=subprocess.DEVNULL,
                       capture_output=True, check=True)
    except subprocess.CalledProcessError as err:
        stdout = err.stdout.decode()
        stderr = err.stderr.decode()
        print()
        msg = "[run_model] Unable to run command (Return code {})"
        print(msg.format(err.returncode))
        print("[run_model] " + " ".join(err.cmd))
        if stdout != "":
            print("[run_model] stdout")
            for line in stdout.split("\n"):
                print(f"\t{line}")
        if stderr != "":
            print("[run_model] stderr")
            for line in stderr.split("\n"):
                print(f"\t{line}")
        raise


def write_data_to_txt(data, tc, fpath, root, fmt):
    """
    Private function, write data to text file.
    """
    # Thread count checker
    if (tc - int(tc) != 0):
        sys.exit('Fit: Invalid tc input - exiting process')

    # Writing data for predictions (if true) or training (if false)
    if root in ["xp", "fp", "xw"]:
        nh = tc
        int_added = 0
    else:
        nh = tc - 1
        int_added = 1
    
    # Check splitted data
    splitted_data = np.array_split(data, nh)
    
    # save to text (accounting for thread number when needed with int_added)
    # Use int_added = 0 when writing prediction grid
    for i, ch in enumerate(splitted_data):
        np.savetxt(str(fpath / Path(root + str(i + int_added))),ch, fmt=fmt)



# Need to generalize -- this is only used in fit
def write_config_file(run_params, fpath, tag):
    """
    UPDATE
    """
    configfile = Path(fpath / ("config"+tag) )
    with configfile.open("w") as tfile:
        for param in run_params:
            tfile.write(str(param) + "\n")


def write_train_data(fpath, xi, x_train, y_train, f_train = None, s_train = None, tc = 2):
    """
    Private function, write data to textfiles.
    """
    # print("splits =", splits)
    n = y_train.shape[0]
    write_data_to_txt(y_train, tc,fpath,"y", '%.7f')
    write_data_to_txt(np.transpose(x_train), tc, fpath, "x", '%.7f')
    write_data_to_txt(np.ones(n, dtype="int"),tc, fpath, "s", '%.0f')
    #print("Results stored in temporary path: " + str(self.fpath))
    if x_train.shape[0] == 1:
        np.savetxt(str(fpath / Path("chgv")), [1], fmt='%.7f')
    elif x_train.shape[0] == 2:
        spr = spearmanr(x_train, axis=1)[0]
        np.savetxt(str(fpath / Path("chgv")),[1,spr,spr,1], fmt='%.7f')
    else:
        np.savetxt(str(fpath / Path("chgv")),spearmanr(x_train, axis=1)[0], fmt='%.7f')

    for k, v in xi.items():
        np.savetxt(str(fpath / Path("xi" + str(k + 1))), v, fmt='%.7f')

    # Write model mixing files
    if not (f_train is None):
        # F-hat matrix
        write_data_to_txt(f_train, tc, fpath, "f", '%.7f')
        # S-hat matrix when using inform_prior
        if not (s_train is None):
            write_data_to_txt(s_train, tc, fpath, "fsd", '%.7f')
        # Wts prior when passed in
        #if self.diffwtsprior:
        #    np.savetxt(str(self.fpath / Path(self.wproot)),
        #                np.concatenate(self.betavec, self.tauvec), fmt='%.7f')





import numpy as np
from scipy.special import factorial

class polynomal_model():
    '''
        Polynomial models class. Used to define a function of the form

        .. math::
            f(x) = c(x-a)^p + b
    '''

    def __init__(self, a=0, b=0, c=1, p=1):
        '''
            Parameters:
            -----------
            :param float a: center parameter.
            :param float b: shift parameter.
            :param float c: scale parameter.
            :param float p: power parameter.

        '''
        self.a = a
        self.b = b
        self.c = c
        self.p = p

    def evaluate(self, x):
        '''
        Evaluate the polynomial at a grid of x's. The standard deviation
        output is set to 1 by default.

        Parameters:
        -----------
        :param np.ndarray x: design matrix.

        Returns:
        --------
        :returns: mean and standard deviation of the model at the x grid points.
        :rtype: np.ndarray, np.ndarray
        :return values: mean predictions.
        :return values: standard deviation of the predictions.

        '''
        if isinstance(x, list):
            x = np.array(x)
        m = self.c * (x - self.a)**self.p + self.b
        if len(m.shape) == 1:
            m = m.reshape(m.shape[0], 1)
        s = np.array([1] * x.shape[0]).reshape(m.shape[0], 1)
        return m, s


# Taylor Expansions
class sin_exp():
    '''
        Sine Taylor series expansion model class.
    '''

    def __init__(self, k, x0):
        '''
        Parameters:
        -----------
        :param int k: the degree of the expansion.
        :param float x0: the center of the expansion.

        '''

        self.k = k
        self.x0 = x0

    def evaluate(self, x):
        '''
        Evaluate the Taylor Series at a grid of x's. The standard deviation
        output is set to 1 by default.

        Parameters:
        -----------
        :param np.ndarray x: design matrix.

        Returns:
        --------
        :returns: mean and standard deviation of the model at the x grid points.
        :rtype: np.ndarray, np.ndarray
        :return values: mean predictions.
        :return values: standard deviation of the predictions.

        '''
        # Check type of x
        if isinstance(x, list):
            x = np.array(x)

        # Get degree list for polynomial expansion
        deg = np.linspace(0, self.k, self.k + 1)

        # Get every 4th term (derivative repaets every 4 terms)
        h0 = deg[deg % 4 == 0]
        h1 = deg[deg % 4 == 1]
        h2 = deg[deg % 4 == 2]
        h3 = deg[deg % 4 == 3]

        # Compute taylor series:
        xc = x - self.x0
        xc0 = xc.repeat(h0.shape[0]).reshape(x.shape[0], h0.shape[0])
        ts = np.sum(
            np.sin(
                self.x0) *
            np.power(
                xc0,
                h0) /
            factorial(h0),
            axis=1)

        xc1 = xc.repeat(h1.shape[0]).reshape(x.shape[0], h1.shape[0])
        ts = ts + np.sum(np.cos(self.x0) * np.power(xc1,
                         h1) / factorial(h1), axis=1)

        xc2 = xc.repeat(h2.shape[0]).reshape(x.shape[0], h2.shape[0])
        ts = ts + np.sum(-np.sin(self.x0) *
                         np.power(xc2, h2) / factorial(h2), axis=1)

        xc3 = xc.repeat(h3.shape[0]).reshape(x.shape[0], h3.shape[0])
        ts = ts + np.sum(-np.cos(self.x0) *
                         np.power(xc3, h3) / factorial(h3), axis=1)

        if len(ts.shape) == 1:
            ts = ts.reshape(ts.shape[0], 1)

        s = np.array([1] * x.shape[0]).reshape(ts.shape[0], 1)
        return ts, s



class cos_exp():
    '''
        Cosine Taylor series expansion model class.
    '''

    def __init__(self, k, x0):
        '''
        Parameters:
        -----------
        :param int k: the degree of the expansion.
        :param float x0: the center of the expansion.

        '''
        self.k = k
        self.x0 = x0

    def evaluate(self, x):
        '''
        Evaluate the Taylor series at a grid of x's. The standard deviation
        output is set to 1 by default.

        Parameters:
        -----------
        :param np.ndarray x: design matrix.

        Returns:
        --------
        :returns: mean and standard deviation of the model at the x grid points.
        :rtype: np.ndarray, np.ndarray
        :return values: mean predictions.
        :return values: standard deviation of the predictions.

        '''
        # Check type of x
        if isinstance(x, list):
            x = np.array(x)

        # Get degree list for polynomial expansion
        deg = np.linspace(0, self.k, self.k + 1)

        # Get every 4th term (derivative repaets every 4 terms)
        h0 = deg[deg % 4 == 0]
        h1 = deg[deg % 4 == 1]
        h2 = deg[deg % 4 == 2]
        h3 = deg[deg % 4 == 3]

        # Compute taylor series:
        xc = x - self.x0
        xc0 = xc.repeat(h0.shape[0]).reshape(x.shape[0], h0.shape[0])
        ts = np.sum(
            np.cos(
                self.x0) *
            np.power(
                xc0,
                h0) /
            factorial(h0),
            axis=1)

        xc1 = xc.repeat(h1.shape[0]).reshape(x.shape[0], h1.shape[0])
        ts = ts + np.sum(-np.sin(self.x0) *
                         np.power(xc1, h1) / factorial(h1), axis=1)

        xc2 = xc.repeat(h2.shape[0]).reshape(x.shape[0], h2.shape[0])
        ts = ts + np.sum(-np.cos(self.x0) *
                         np.power(xc2, h2) / factorial(h2), axis=1)

        xc3 = xc.repeat(h3.shape[0]).reshape(x.shape[0], h3.shape[0])
        ts = ts + np.sum(np.sin(self.x0) * np.power(xc3,
                         h3) / factorial(h3), axis=1)

        if len(ts.shape) == 1:
            ts = ts.reshape(ts.shape[0], 1)

        s = np.array([1] * x.shape[0]).reshape(ts.shape[0], 1)
        return ts, s


class sin_cos_exp():
    '''
    Taylor series expansion of

    .. math::
            f(x) = \\sin(x_1) + \\cos(x_2)


    '''

    def __init__(self, ks, kc, xs, xc):
        '''
        Parameters:
        -----------
        :param int ks: the degree of the sine expansion.
        :param int kc: the degree of the cosine expansion.
        :param float xs: the center of the sine expansion.
        :param float xc: the center of the cosine expansion.

        '''

        self.ks = ks
        self.xs = xs

        self.kc = kc
        self.xc = xc

    def evaluate(self, x):
        '''
        Evaluate the model at a grid of x's. The standard deviation
        output is set to 1 by default.

        Parameters:
        -----------
        :param np.ndarray x: design matrix.

        Returns:
        --------
        :returns: mean and standard deviation of the model at the x grid points.
        :rtype: np.ndarray, np.ndarray
        :return values: mean predictions.
        :return values: standard deviation of the predictions.

        '''
        # Check type of x
        if isinstance(x, list):
            x = np.array(x)

        # Sine Part
        # Get degree list for polynomial expansion
        deg = np.linspace(0, self.ks, self.ks + 1)

        # Get every 4th term (derivative repaets every 4 terms)
        h0 = deg[deg % 4 == 0]
        h1 = deg[deg % 4 == 1]
        h2 = deg[deg % 4 == 2]
        h3 = deg[deg % 4 == 3]

        # Compute taylor series:
        xctr = x.transpose()[0] - self.xs
        xc0 = xctr.repeat(h0.shape[0]).reshape(x.shape[0], h0.shape[0])
        tss = np.sum(
            np.sin(
                self.xs) *
            np.power(
                xc0,
                h0) /
            factorial(h0),
            axis=1)

        xc1 = xctr.repeat(h1.shape[0]).reshape(x.shape[0], h1.shape[0])
        tss = tss + np.sum(np.cos(self.xs) * np.power(xc1,
                           h1) / factorial(h1), axis=1)

        xc2 = xctr.repeat(h2.shape[0]).reshape(x.shape[0], h2.shape[0])
        tss = tss + np.sum(-np.sin(self.xs) *
                           np.power(xc2, h2) / factorial(h2), axis=1)

        xc3 = xctr.repeat(h3.shape[0]).reshape(x.shape[0], h3.shape[0])
        tss = tss + np.sum(-np.cos(self.xs) *
                           np.power(xc3, h3) / factorial(h3), axis=1)

        # Cosine Part
        # Get degree list for polynomial expansion
        deg = np.linspace(0, self.kc, self.kc + 1)

        # Get every 4th term (derivative repaets every 4 terms)
        h0 = deg[deg % 4 == 0]
        h1 = deg[deg % 4 == 1]
        h2 = deg[deg % 4 == 2]
        h3 = deg[deg % 4 == 3]

        # Compute taylor series:
        xctr = x.transpose()[1] - self.xc
        xc0 = xctr.repeat(h0.shape[0]).reshape(x.shape[0], h0.shape[0])
        tsc = np.sum(
            np.cos(
                self.xc) *
            np.power(
                xc0,
                h0) /
            factorial(h0),
            axis=1)

        xc1 = xctr.repeat(h1.shape[0]).reshape(x.shape[0], h1.shape[0])
        tsc = tsc + np.sum(-np.sin(self.xc) *
                           np.power(xc1, h1) / factorial(h1), axis=1)

        xc2 = xctr.repeat(h2.shape[0]).reshape(x.shape[0], h2.shape[0])
        tsc = tsc + np.sum(-np.cos(self.xc) *
                           np.power(xc2, h2) / factorial(h2), axis=1)

        xc3 = xctr.repeat(h3.shape[0]).reshape(x.shape[0], h3.shape[0])
        tsc = tsc + np.sum(np.sin(self.xc) * np.power(xc3,
                           h3) / factorial(h3), axis=1)

        # Add the sine and cosine parts
        ts = tss + tsc

        if len(ts.shape) == 1:
            ts = ts.reshape(ts.shape[0], 1)

        s = np.array([1] * x.shape[0]).reshape(ts.shape[0], 1)
        return ts, s



from logging import raiseExceptions
import numpy as np
import matplotlib.pyplot as plt
import sys
import subprocess
import tempfile
import shutil
import os
import typing


from pathlib import Path

import openbtmixing.interface_helpers as ih

#from importlib import reload
#reload(ih)

class Openbtmix():
    def __init__(self,**kwargs):
        '''
        .. todo::
            * See if we can remove local_openbt_path and google_colab

        Parameters:
        ----------
        :param dict model_dict:
            Dictionary of models where each item is an instance of BaseModel.

        :param dict kwargs:
            Additional arguments to pass to the constructor.

        Returns:
        ---------
        :returns: None.
        '''

        # MCMC Parameters
        self.ndpost = 1000
        self.nskip = 100
        self.nadapt = 1000
        self.tc = 2
        self.pbd = 0.7
        self.pb = 0.5
        self.stepwpert = 0.1
        self.probchv = 0.1
        self.minnumbot = 5
        self.printevery = 100
        self.numcut = 100
        self.adaptevery = 100

        # Cutpoints
        self.xicuts = None

        # Set the prior defaults
        self.nummodels = 1
        self.sighat = 1
        self.nu = 10
        self.k = 2
        self.ntree = 1
        self.ntreeh = 1
        self.power = 2.0
        self.base = 0.95
        self.inform_prior = False

        # Set defaults for variance product trees, assume constant variance in this model
        self.baseh = 0.95
        self.powerh = 2.0
        self.minnumboth = 5
        self.pbdh = 0
        self.pbh = 0
        self.stepwperth = 0.1
        self.probchvh = 0

        # Define the roots for the output files --- might be able to delete after hard code
        self.xroot = "x"
        self.yroot = "y"
        self.sroot = "s"
        self.chgvroot = "chgv"
        self.froot = "f"
        self.fsdroot = "fsd"
        self.wproot = "wpr"
        self.xiroot = "xi"

        # Set other defaults
        self.modelname = "mixmodel"
        self.summarystats = "FALSE"
        self.local_openbt_path = ""
        self.google_colab = False       
        self.diffwtsprior = False # used for diff values of beta and tau in noninform prior
        self.tauvec = None # the tauvec for diff components in noninform prior
        self.betavec = None # the betavec for diff components in noninform prior

        # Set the kwargs dictionary
        self.__dict__.update((key, value) for key, value in kwargs.items())

        # Set defaults not relevant to model mixing -- only set so cpp code
        # doesn't break
        self.truncateds = None
        self.modeltype = 9
        self._is_prior_set = False
        self._is_predict_run = False  # Remove ????

        # Setter for prior to compute tau, beta, ect
        self.set_prior(self.ntree,self.ntreeh,self.k,self.power,self.base,
                       self.sighat,self.nu,self.inform_prior)


    def get_prior(self):
        out = {}
        out["beta"] = self.beta
        out["tau"] = self.tau
        out["nu"] = self.nu
        out["lam"] = self.lam
        out["base"] = self.base
        out["power"] = self.power
        return out


    def get_mcmc_info(self):
        out = {}
        out["pbd"] = self.pbd
        out["pb"] = self.pb
        out["stepwpert"] = self.stepwpert
        out["probchv"] = self.probchv
        out["nadapt"] = self.nadapt
        out["adaptevery"] = self.adaptevery
        out["nskip"] = self.nskip
        out["ndpost"] = self.ndpost
        return out

    
    
    def set_cutpoints(self, xmax, xmin, numcut):
        self.xi = {}
        self.numcut = numcut
        for j in range(len(xmax)):
            xinc = (xmax[j] - xmin[j]) / (self.numcut + 1)
            self.xi[j] = [np.arange(1, (numcut) + 1) * xinc + xmin[j]]


    def set_mcmc_info(self, mcmc_dict):
        # Extract arguments
        valid_mcmc_args = [
            "ndpost",
            "nskip",
            "nadapt",
            "adaptevery",
            "tc",
            "minnumbot",
            "printevery"]
        
        for (key, value) in mcmc_dict.items():
            if key in valid_mcmc_args:
                self.__dict__.update({key: value})
            else:
                print((key,value))


    def set_prior(
            self,
            ntree: int = 1,
            ntreeh: int = 1,
            k: float = 2,
            power: float = 2.0,
            base: float = 0.95,
            sighat: float = 1,
            nu: int = 10,
            inform_prior: bool = True):
        '''
        Sets the hyperparameters in the tree and terminal node priors. Also
        specifies if an informative or non-informative prior will be used when mixing EFTs.

        Parameters:
        -----------
        :param int ntree:
            The number of trees used in the sum-of-trees model for
            the weights.
        :param int ntreeh:
            The number of trees used in the product-of-trees model
            for the error standard deviation. Set to 1 for
            homoscedastic variance assumption.
        :param float k:
            The tuning parameter in the prior variance of the terminal node
            parameter prior. This is a value greater than zero.
        :param float power:
            The power parameter in the tree prior.
        :param float base:
            The base parameter in the tree prior.
        :param float overallsd:
            An initial estimate of the erorr standard deviation.
            This value is used to calibrate the scale parameter in variance prior.
        :param float overallnu:
            The shape parameter in the error variance prior.
        :param bool inform_prior:
            Controls if the informative or non-informative prior is used.
            Specify true for the informative prior.

        Returns:
        --------
        :returns: None.

        '''
        # Extract arguments
        # valid_prior_args = ['ntree', 'ntreeh', 'k','power','base','overallsd','overallnu','inform_prior','tauvec', 'betavec']
        prior_dict = {
            'ntree': ntree,
            'ntreeh': ntreeh,
            'k': k,
            'power': power,
            'base': base,
            'sighat': sighat,
            'nu': nu,
            'inform_prior': inform_prior}

        self._prior = dict()
        for (key, value) in prior_dict.items():
            # if key in valid_prior_args:
            # Update class dictionary which stores all objects
            self.__dict__.update({key: value})
            # Updates prior specific dictionary
            self._prior.update({key: value})


        # Set prior for terminal node parameters
        # Default non-informative mean is 1/(2*m)
        # for 2 models. This gets updated in train()
        # if more than 2 models
        if self.inform_prior:
            self.tau = 1 / (2*self.ntree*self.k)
            self.beta = 1 / self.ntree
        else:
            self.tau = 1 / (2*np.sqrt(self.ntree)*self.k)
            self.beta = 1 / (2*self.ntree) 

        # Set lambda and nu
        self.nu = nu
        self.lam = ((nu + 2)/nu)*sighat*sighat
        

    
    def train(self, x_train: np.ndarray, y_train: np.ndarray, f_train: np.ndarray,
              s_train = None, **kwargs):
        '''
        Train the mixed-model using a set of observations y at inputs x.

        Parameters:
        ----------
        :param np.ndarray X: input parameter values of dimension (n x p).
        :param np.ndarray y: observed data at inputs X of dimension  (n x 1).
        :param dict kwargs: dictionary of arguments

        Returns:
        --------
        :returns: A dictionary which contains relevant information to the model such as
            values of tuning parameters. The MCMC results are written to a text file
            and stored in a temporary directory as defined by the fpath key in the
            results dictionary.
        :rtype: dict

        '''
        # Cast data to arrays if not already and reshape if needed
        if isinstance(x_train, list):
            x_train = np.array(x_train)

        if len(x_train.shape) == 1:
            x_train = x_train.reshape(x_train.shape[0], 1)

        if isinstance(y_train, list):
            y_train = np.array(y_train)

        if len(y_train.shape) == 1:
            y_train = y_train.reshape(y_train.shape[0], 1)

        # Dimension Checks
        if x_train.shape[0] != y_train.shape[0]:
            raise ValueError("Number of rows in x_train does not match length of y_train.")

        if x_train.shape[0] != f_train.shape[0]:
            raise ValueError("Number of rows in x_train does not match length of f_train.")
        
        if self.inform_prior:
            if f_train.shape != s_train.shape:
                raise ValueError("Shape of f_train does not match shape of s_train.")
        
        # Store n and p
        n = x_train.shape[0]
        self.p = x_train.shape[1]

        # Transpose x
        x_train = np.transpose(x_train)

        # Set default outputs
        self.fmean_out = 0
        
        # Cutpoints
        if self.__dict__["xicuts"] is None:
            # Get the bounds
            xmax = np.ceil(np.max(x_train, axis=1))
            xmin = np.floor(np.min(x_train, axis=1))
            if "numcut" in kwargs.keys():
                self.numcut = kwargs["numcut"]    
            self.set_cutpoints(xmax, xmin, self.numcut)


        # Set the mcmc properties from kwargs
        self.set_mcmc_info(kwargs)

        # Create path
        f = tempfile.mkdtemp(prefix="openbtmixing_")
        self.fpath = Path(f)

        # Update beta if number of models changes and using informative prior
        self.nummodels = f_train.shape[1]
        if not self.inform_prior and (self.nummodels > 2):
            self.beta = 1 / (self.nummodels*self.ntree)

        # Get config parameters
        run_params = [self.modeltype,self.xroot,self.yroot,self.fmean_out,self.ntree,self.ntreeh,
                        self.ndpost,self.nskip,self.nadapt,self.adaptevery,self.tau,self.beta,
                        self.lam,self.nu,self.base,self.power,self.baseh,self.powerh,self.tc,
                        self.sroot,self.chgvroot,self.froot,self.fsdroot,self.inform_prior,
                        self.wproot,self.diffwtsprior,self.pbd,self.pb,self.pbdh,self.pbh,
                        self.stepwpert,self.stepwperth,self.probchv,self.probchvh,self.minnumbot,
                        self.minnumboth,self.printevery,self.xiroot,self.modelname,self.summarystats]


        # Write config file
        ih.write_config_file(run_params, self.fpath, tag = "")
        if self.inform_prior:
            ih.write_train_data(self.fpath, self.xi, x_train, y_train, f_train, s_train = s_train, tc = self.tc)
        else:
            ih.write_train_data(self.fpath, self.xi, x_train, y_train, f_train, s_train = None, tc = self.tc)

        print("Running model...")        
        cmd = "openbtcli"
        ih.run_model(self.fpath, self.tc, cmd, local_openbt_path = self.local_openbt_path, google_colab = self.google_colab)

        # See which attributes are returned/can be returned here
        # Return attributes to be saved as a separate fit object:
        # Missing the influence attribute from the R code (skip for now)
        res = {}
        keys = ['fpath','inform_prior','minnumbot','nummodels','p']

        for key in keys:
            res[key] = self.__dict__[key]

        return res

    
    def predict(self, x_test: np.ndarray, f_test: np.ndarray, ci: float = 0.95):
        '''
        Obtain the posterior predictive distribution of the mixed-model at a set
        of inputs X.

        Parameters:
        ----------
        :param np.ndarray X: design matrix of testing inputs.
        :param float ci: credible interval width, must be a value within the interval (0,1).

        Returns:
        --------
        :returns: The posterior prediction draws and summaries.
        :rtype: np.ndarray, np.ndarray, np.ndarray, np.ndarray
        :return value: the posterior predictive distribution evaluated at the specified test points
        :return value: the posterior mean of the mixed-model at each input in X.
        :return value: the pointwise credible intervals at each input in X.
        :return value: the posterior standard deviation of the mixed-model at each input in X.
        '''

        # Set q_lower and q_upper
        alpha = (1 - ci)
        q_lower = alpha / 2
        q_upper = 1 - alpha / 2

        # Casting lists to arrays when needed
        if (isinstance(x_test, list)):
            x_test = np.array(x_test)
        if (len(x_test.shape) == 1):  # If shape is (n, ), change it to (n, 1):
            x_test = x_test.reshape(len(x_test), 1)


        # Set control values
        p_test = x_test.shape[1]

        if p_test != self.p:
            raise ValueError("Number of inputs in x_test does not match number of inputs in x_train.")
        
        if f_test.shape[1] != self.nummodels:
            raise ValueError("Number of models in f_test does not equal number of models in f_train.")

        self.xproot = "xp"
        self.fproot = "fp"

        ih.write_data_to_txt(x_test, self.tc, self.fpath, self.xproot,'%.7f')
        ih.write_data_to_txt(f_test, self.tc, self.fpath, self.fproot,'%.7f')

        # Set and write config file
        pred_params = [self.modelname, self.modeltype,
                       self.xiroot, self.xproot, self.fproot,
                       self.ndpost, self.ntree, self.ntreeh,
                       p_test, self.nummodels, self.tc, self.fmean_out]
        ih.write_config_file(pred_params, self.fpath, tag = ".pred")

        # Run prediction
        cmd = "openbtpred"
        ih.run_model(fpath = self.fpath, tc = self.tc, cmd = cmd, local_openbt_path = self.local_openbt_path, google_colab = self.google_colab)
        
        res = ih.read_in_preds(self.fpath, self.modelname, self.nummodels, q_upper, q_lower, True, True, False)        
        return res

    
    def predict_weights(self, x_test: np.ndarray, ci: float = 0.95):
        '''
        Obtain posterior distribution of the weight functions at a set
        of inputs X.

        Parameters:
        ----------
        :param np.ndarray X: design matrix of testing inputs.
        :param float ci: credible interval width, must be a value within the interval (0,1).

        Returns:
        --------
        :returns: The posterior weight function draws and summaries.
        :rtype: np.ndarray, np.ndarray, np.ndarray, np.ndarray
        :return value: the posterior draws of the weight functions at each input in X.
        :return value: posterior mean of the weight functions at each input in X.
        :return value: pointwise credible intervals for the weight functions.
        :return value: posterior standard deviation of the weight functions at each input in X.
        '''

        # Set q_lower and q_upper
        alpha = (1 - ci)
        q_lower = alpha / 2
        q_upper = 1 - alpha / 2

        # Checks for proper inputs and convert lists to arrays
        if not self.modeltype == 9:
            raise TypeError("Cannot call openbt.mixingwts() method for openbt objects that are not modeltype = 'mixbart'")
        if isinstance(x_test, list):
            x_test = np.array(x_test)
        if (len(x_test.shape) == 1):
            x_test = x_test.reshape(len(x_test), 1)

        # Set control values
        p_test = x_test.shape[1]
        if p_test != self.p:
            raise ValueError("Number of inputs in x_test does not match number of inputs in x_train.")


        # Set control parameters
        self.xwroot = "xw"
        # default, needed when considering the general class of model mixing
        # problems -- revist this later
        self.fitroot = ".fit"

        # write out the config file
        # Set control values
        # no need to set this as a class object like in predict function
        ih.write_data_to_txt(x_test, self.tc, self.fpath, self.xwroot,'%.7f')

        # Set and write config file
        pred_params = [self.modelname, self.modeltype,
                       self.xiroot, self.xwroot, self.fitroot,
                       self.ndpost, self.ntree, self.ntreeh,
                       p_test, self.nummodels, self.tc]
        
        ih.write_config_file(pred_params, self.fpath, tag = ".mxwts")
        
        # run the program
        cmd = "openbtmixingwts"
        ih.run_model(self.fpath, self.tc, cmd, local_openbt_path = self.local_openbt_path, google_colab = self.google_colab)

        res = ih.read_in_preds(self.fpath, self.modelname, self.nummodels, q_upper, q_lower, False, False, True)        
        return res




"""
Name: test_trees.py
Author: John Yannotty (yannotty.1@osu.edu)
Desc: Test suite for trees.py

Start Date: 04/21/22
Version: 1.0

"""
# Imports
import os
import sys
import numpy as np

from pathlib import Path

from openbtmixing.tests.polynomial_models import sin_cos_exp
from openbtmixing import Openbtmix


_TEST_DATA = Path(__file__).parent.joinpath("bart_bmm_test_data").resolve()


# ---------------------------------------------
# Define the test functions
# ---------------------------------------------
# Test the mixing fun
def test_mixing():
    x_train = np.loadtxt(_TEST_DATA.joinpath('2d_x_train.txt')).reshape(80, 2)
    x_train = x_train.reshape(2, 80).transpose()

    y_train = np.loadtxt(_TEST_DATA.joinpath('2d_y_train.txt')).reshape(80, 1)

    f_train = np.concatenate([f1.evaluate(x_train)[0],f2.evaluate(x_train)[0]], axis = 1)

    # Set prior information
    sighat = 0.01/np.sqrt(7/5)
    mix.set_prior(
        k=2.5,
        ntree=30,
        nu=5,
        sighat=sighat,
        inform_prior=False)

    # Check tuning & hyper parameters
    assert mix.k == 2.5, "class object k is not set."
    assert mix.ntree == 30, "class object ntree is not set."
    assert mix.nu == 5, "class object nu is not set."
    assert mix.lam == 0.01**2, "class object lambda is not set."
    assert mix.inform_prior == False, "class object inform_prior is not set."

    # Train the model
    #
    # The GitHub action runners can have as few as two processors.  When tests
    # run on those with Open MPI with more MPI processes than processors, it
    # exits due to oversubscription.  The value of tc is set to get all Open
    # MPI-based actions running.
    fit = mix.train(
        x_train=x_train,
        y_train=y_train,
        f_train=f_train,
        ndpost=10000,
        nadapt=2000,
        nskip=2000,
        adaptevery=500,
        minnumbot=4,
        tc = 2)

    # Check the mcmc objects
    assert mix.ndpost == 10000, "class object ndpost is not set."
    assert mix.nadapt == 2000, "class object nadapt is not set."
    assert mix.adaptevery == 500, "class object adaptevery is not set."
    assert mix.nskip == 2000, "class object nskip is not set."
    assert mix.minnumbot == 4, "class object minnumbot is not set."
    assert mix.nummodels == 2, "class object nummodels is not set."


# Test the mean predictions
def test_predict():
    # Get test data
    n_test = 30
    x1_test = np.outer(np.linspace(-3, 3, n_test), np.ones(n_test))
    x2_test = x1_test.copy().transpose()
    x_test = np.array([x1_test.reshape(x1_test.size,),
                      x2_test.reshape(x1_test.size,)]).transpose()
    f_test = np.concatenate([f1.evaluate(x_test)[0],f2.evaluate(x_test)[0]], axis = 1)

    # Read in test results
    pmean_test = np.loadtxt(_TEST_DATA.joinpath('2d_pmean.txt'))
    eps = 0.10

    # Get predictions
    res = mix.predict(x_test=x_test,f_test = f_test, ci=0.95)

    # Test the values
    perr = np.mean(np.abs(res["pred"]["mean"] - pmean_test))
    assert perr < eps, "Inaccurate predictions."


# Test posterior of the weights
def test_predict_wts():
    # Get weights
    n_test = 30
    x1_test = np.outer(np.linspace(-3, 3, n_test), np.ones(n_test))
    x2_test = x1_test.copy().transpose()
    x_test = np.array([x1_test.reshape(x1_test.size,),
                      x2_test.reshape(x1_test.size,)]).transpose()

    res = mix.predict_weights(x_test=x_test, ci=0.95)

    # Read in test results
    wteps = 0.05
    wmean_test = np.loadtxt(_TEST_DATA.joinpath('2d_wmean.txt'))

    # Test the values
    werr = np.mean(np.abs(res["wts"]["mean"] - wmean_test))
    assert werr < wteps, "Inaccurate weights."


# Test sigma
def test_sigma():
    sig_eps = 0.05
    n_test = 5
    x1_test = np.outer(np.linspace(-3, 3, n_test), np.ones(n_test))
    x2_test = x1_test.copy().transpose()
    x_test = np.array([x1_test.reshape(x1_test.size,),
                      x2_test.reshape(x1_test.size,)]).transpose()
    f_test = np.concatenate([f1.evaluate(x_test)[0],f2.evaluate(x_test)[0]],axis = 1)

    res = mix.predict(x_test=x_test, f_test = f_test, ci=0.95)
    assert np.abs((res["sigma"]["mean"][0] - 0.1)) < sig_eps, "Inaccurate sigma calculation."


# ---------------------------------------------
# Initiatilize model set
# ---------------------------------------------
# Define the model set
f1 = sin_cos_exp(7, 10, np.pi, np.pi)
f2 = sin_cos_exp(13, 6, -np.pi, -np.pi)

mix = Openbtmix()



import os
import sys
import shutil
import codecs

import subprocess as sbp

from pathlib import Path
from setuptools import (
    setup, Command
)
from setuptools.command.build import build as _build

# ----- HARDCODED VALUES
PKG_ROOT = Path(__file__).resolve().parent
PY_SRC_PATH = PKG_ROOT.joinpath("src", "openbtmixing")
CLT_SRC_PATH = PKG_ROOT.joinpath("cpp")

# Names of C++ products to include
#
# Only a subset of OpenBT command line tools are used in package.  Any changes
# made here should also be made in interface_helpers.py:run_model.
CLT_NAMES = ["openbtcli", "openbtpred", "openbtmixingwts"]

# Package metadata
PYTHON_REQUIRES = ">=3.9"
CODE_REQUIRES = ["setuptools", "numpy", "matplotlib"]
TEST_REQUIRES = ["pytest", "scipy", "pandas"]
INSTALL_REQUIRES = CODE_REQUIRES + TEST_REQUIRES

PACKAGE_DATA = {
    "openbtmixing":
        ["tests/bart_bmm_test_data/2d_*.txt"] +
        [f"bin/{clt}" for clt in CLT_NAMES]
}

PROJECT_URLS = {
    "Source": "https://github.com/jcyannotty/OpenBT",
    "Documentation": "https://github.com/jcyannotty/OpenBT",
    "Tracker": "https://github.com/jcyannotty/OpenBT/issues",
}


# ----- CUSTOM COMMAND TO BUILD C++ CLTs
# EdgeDB builds a CLT that they make available through their interface.  This
# is partially based on what they have done at commit c5db98c.
#
# https://github.com/edgedb/edgedb/blob/master/setup.py
class build(_build):
    sub_commands = ([("build_clt", None)])


class build_clt(Command):
    description = "Build the OpenBTMixing command line tools"

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self, *args, **kwargs):
        if shutil.which("meson", mode=os.F_OK | os.X_OK) is None:
            print()
            print("Please install the Meson build system & add meson to path")
            print()
            sys.exit(1)

        # To build debug versions with more output,
        # * use --buildtype=debug
        # * consider using -Dverbose=true
        # * consider adding arguments such as --warnlevel and --werror to
        #   SETUP_CMD
        # * Remove "--quiet" from INSTALL_CMD
        # * Use python -m pip install -v ...
        SETUP_CMD = ["meson", "setup", "--wipe", "--clearcache",
                     "--buildtype=release", "builddir", f"-Dprefix={PY_SRC_PATH}",
                     "-Duse_mpi=true", "-Dverbose=false", "-Dpypkg=true"]
        COMPILE_CMD = ["meson", "compile", "-v", "-C", "builddir"]
        INSTALL_CMD = ["meson", "install", "--quiet", "-C", "builddir"]

        # Install the CLTs within the Python source files and so that they are
        # included in the wheel build based on PACKAGE_DATA
        cwd = Path.cwd()
        os.chdir(CLT_SRC_PATH)
        for cmd in [SETUP_CMD, COMPILE_CMD, INSTALL_CMD]:
            try:
                sbp.run(cmd, stdin=sbp.DEVNULL, capture_output=False, check=True)
            except sbp.CalledProcessError as err:
                print()
                msg = "[meson build] Unable to run command (Return code {})"
                print(msg.format(err.returncode))
                print("[meson build] " + " ".join(err.cmd))
                sys.exit(2)
        os.chdir(cwd)


cmdclass = {
    'build': build,
    'build_clt': build_clt
}


# ----- SPECIFY THE PACKAGE
def readme_md():
    fname = PKG_ROOT.joinpath("README.md")
    with codecs.open(fname, "r", encoding="utf8") as fptr:
        return fptr.read()


def version():
    fname = PKG_ROOT.joinpath("VERSION")
    with open(fname, "r") as fptr:
        return fptr.read().strip()


setup(
    name='openbtmixing',
    version=version(),
    author="John Yannotty",
    author_email="yannotty.1@buckeyemail.osu.edu",
    maintainer="John Yannotty",
    maintainer_email="yannotty.1@buckeyemail.osu.edu",
    license="MIT",
    package_dir={"": "src"},
    package_data=PACKAGE_DATA,
    cmdclass=cmdclass,
    url=PROJECT_URLS["Source"],
    project_urls=PROJECT_URLS,
    description="Model mixing using Bayesian Additive Regression Trees",
    long_description=readme_md(),
    long_description_content_type="text/markdown",
    python_requires=PYTHON_REQUIRES,
    install_requires=INSTALL_REQUIRES,
    classifiers=[
        "Programming Language :: Python :: 3"
    ]
)
